{ File    : $URL: file:///D:/SVN/svnDelphi/trunk/A%20New%20Beginning/VISExt/src/VISExt.dpr $
  Author  : $Author: Deniz Özmen $
  Changed : $Date: 2012-07-15 18:12:07 +0200 (So, 15 Jul 2012) $
  Revision: $Rev: 61 $

  Visionaire Engine VIS extractor, based on Luigi Auriemma's visionaire.bms.
  Thus:


  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
}
program VISExt;

{$APPTYPE CONSOLE}

{$R '..\res\version.res' '..\res\version.rc'}

uses
  KeyCfg in 'backend\KeyCfg.pas',
  DCRC in 'utils\DCRC.pas',
  MemSearch in 'utils\MemSearch.pas',
  SdcpMD5 in 'utils\SdcpMD5.pas',
  Classes,
  StrUtils,
  SysUtils,
  ZLib;

const
  VISHeaderID = 'VIS3';
  DirHeaderID = 'HDR';
  DirFooterID = 'END';

  Flag_MainXML1     = $00000001;
  Flag_CryptDefault = $00000002;
  Flag_CryptPNG     = $00000008;
  Flag_MainXML2     = $00000010;

  Flag_MainXML      = Flag_MainXML1 or Flag_MainXML2;

  CompChainEnd: Longword = $ffeeffee;

type
  TParamRec = record
    FileSpec: AnsiString;
    NameSpec: AnsiString;
    GenNames: Boolean;
    UseNames: Boolean;
    Unique  : Boolean;
  end;

  TVISHeader = packed record
    ID      : array[0..3] of Char;
    NumFiles: Longword;
  end;

  TDirHeader = packed record
    ID: array[0..2] of Char;
  end;

  TDirFooter = packed record
    ID: array[0..2] of Char;
  end;

  TDirEntry = packed record
    Offset    : Longword;
    CompSize  : Longword;
    UncompSize: Longword;
    Flags     : Longword;
  end;

  TDirEntries = packed array of TDirEntry;

  TCompBlockHeader = packed record
    UncompSize: Longword;
    CompSize  : Longword;
  end;

  TBuffer = array of Byte;

var
  ParamRec : TParamRec;
  GameIdx  : Longint;
  NameTable: array of AnsiString;
  BigEndian: Boolean;

function OpenInFile(var InFile: file; InName: string): Boolean;
begin
  FileMode := fmShareDenyWrite;
  AssignFile(InFile, InName);

  {$I-}
  Reset(InFile, 1);
  {$I+}
  Result := (IOResult = 0)
end;

function OpenOutFile(var OutFile: file; OutName: string): Boolean;
var
  Dir: AnsiString;
begin
  // first create the directory portion of the output file
  Dir := ExtractFileDir(ExpandFileName(OutName));
  ForceDirectories(Dir);

  // open the output file
  FileMode := fmOpenWrite;
  AssignFile(OutFile, OutName);

  {$I-}
  Rewrite(OutFile, 1);
  {$I+}
  Result := (IOResult = 0)
end;

procedure SwitchEndian(var Val: Longword; Force: Boolean = False);
begin
  if BigEndian or Force then
    Val := (Val shr 24) or
           ((Val shr 8) and $0000ff00) or
           ((Val shl 8) and $00ff0000) or
           (Val shl 24)
end;

function ReadHeader(var VISFile: file; out VISHeader: TVISHeader): Boolean;
var
  BytesRead: Longint;
  NumFiles : Longword;
begin
  BlockRead(VISFile, VISHeader, SizeOf(VISHeader), BytesRead);

  NumFiles := VISHeader.NumFiles;

  BigEndian := True;
  SwitchEndian(VISHeader.NumFiles);

  if NumFiles < VISHeader.NumFiles then
  begin
    BigEndian := False;
    VISHeader.NumFiles := NumFiles
  end;

  Result := (SizeOf(VISHeader) = BytesRead) and (VISHeader.ID = VISHeaderID)
end;

function Decompress(const InBuffer: TBuffer; InOfs, InLen: Longint;
  var OutBuffer: TBuffer; OutOfs, OutLen: Longint): Boolean;
var
  DecompSize: Longint;
  TempBuffer: Pointer;
begin
  // standard ZLib decompression
  DecompressBuf(@InBuffer[InOfs], InLen, OutLen, TempBuffer,
    DecompSize);

  Result := (DecompSize = OutLen);
  if Result then
    Move(TempBuffer^, OutBuffer[OutOfs], DecompSize);

  FreeMem(TempBuffer, DecompSize)
end;

procedure Decrypt(const Buffer: TBuffer; Offset, Len: Longint;
  const Key: AnsiString);
var
  BufIdx,
  KeyIdx: Longint;
begin
  KeyIdx := 1;

  for BufIdx := Offset to Offset + Len - 1 do
  begin
    if (BufIdx < Low(Buffer)) or (BufIdx > High(Buffer)) then
      Break;

    Buffer[BufIdx] := Buffer[BufIdx] xor Ord(Key[KeyIdx]);

    KeyIdx := KeyIdx mod Length(Key) + 1
  end
end;

function DecryptPNG(const Buffer: TBuffer; PNGName: AnsiString): Boolean;
type
  TPNG_IHDR = packed record
    ID             : array[0..3] of Char;
    Width          : Longint;
    Height         : Longint;
    BitDepth       : Byte;
    ColorType      : Byte;
    CompMethod     : Byte;
    FilterMethod   : Byte;
    InterlaceMethod: Byte;
  end;
var
  Pattern  : array[0..3] of Char;
  Idx      : Longint;
  TestBuf  : TBuffer;
  WideName : WideString;
  MD5      : TDCP_md5;
  Digest   : array[0..3] of Longword;
  Key      : AnsiString;
  CRCVal,
  TargetCRC: Longword;

  function ValidatePNG: Boolean;
  begin
    // try the decryption ...
    SwitchEndian(Digest[0], True);
    Key := AnsiLowerCase(Format('%.8x', [Digest[0]]));
    TestBuf[4]  := TestBuf[4]  xor Ord(Key[1]);
    TestBuf[5]  := TestBuf[5]  xor Ord(Key[2]);
    TestBuf[6]  := TestBuf[6]  xor Ord(Key[3]);
    TestBuf[7]  := TestBuf[7]  xor Ord(Key[4]);
    TestBuf[8]  := TestBuf[8]  xor Ord(Key[5]);
    TestBuf[9]  := TestBuf[9]  xor Ord(Key[6]);
    TestBuf[10] := TestBuf[10] xor Ord(Key[7]);
    TestBuf[11] := TestBuf[11] xor Ord(Key[8]);

    // ... and validate the chunk checksum
    CRCVal := $ffffffff;
    CalcCRC32(@TestBuf[0], Length(TestBuf), CRCVal);
    CRCVal := not CRCVal;

    // if the key was correct, the test buffer contains the decrypted data
    Result := (CRCVal = TargetCRC)
  end;

begin
  // The width and height values of the PNG files stored in the archives of "A
  // New Beginning" have been encrypted. The encryption method is basically
  // the same as the one used to obfuscate the file table structure, but the
  // encryption key varies from file to file based on the file name.
  Result := False;

  if Length(Buffer) < SizeOf(TPNG_IHDR) + SizeOf(TargetCRC) then
  begin
    WriteLn('    warning: data too short for a PNG file');
    Exit
  end;

  // search the IHDR chunk
  Pattern := 'IHDR';
  Idx := SearchMem(SearchMethod_KnuthMorrisPratt, @Pattern[0], @Buffer[0],
    Length(Pattern), Length(Buffer));

  if (Idx < 0) or (Idx + SizeOf(TPNG_IHDR) + SizeOf(TargetCRC) >
    Length(Buffer)) then
  begin
    WriteLn('    warning: no valid IHDR found');
    Exit
  end;

  // get the targetted CRC value of the IHDR chunk
  Move(Buffer[Idx + SizeOf(TPNG_IHDR)], TargetCRC, SizeOf(TargetCRC));
  SwitchEndian(TargetCRC, True);

  // compute the decryption key from the prefix part of the file name
  PNGName := ExtractFileName(PNGName);
  PNGName := Copy(PNGName, 1, Length(PNGName) -
    Length(ExtractFileExt(PNGName)));

  MD5 := TDCP_md5.Create;

  // first try: use the prefix as a regular ANSI string
  SetLength(TestBuf, SizeOf(TPNG_IHDR));
  Move(Buffer[Idx], TestBuf[0], Length(TestBuf));

  MD5.Init;
  MD5.Update(PNGName[1], Length(PNGName));
  MD5.Final(Digest);

  Result := ValidatePNG;
  if not Result then
  begin
    // if the ANSI string failed, try the wide variant of the prefix
    SetLength(TestBuf, SizeOf(TPNG_IHDR));
    Move(Buffer[Idx], TestBuf[0], Length(TestBuf));

    WideName := PNGName;
    MD5.Init;
    MD5.Update(WideName[1], Length(WideName) * 2);
    MD5.Final(Digest);

    Result := ValidatePNG;
  end;

  if Result then
    Move(TestBuf[0], Buffer[Idx], Length(TestBuf))
  else
    WriteLn('    warning: no valid PNG encryption key derivable from file nam' +
      'e');

  MD5.Free
end;

function DecryptPNGForce(const Buffer: TBuffer): Boolean;
type
  TPNG_IHDR = packed record
    ID             : array[0..3] of Char;
    Width          : Longint;
    Height         : Longint;
    BitDepth       : Byte;
    ColorType      : Byte;
    CompMethod     : Byte;
    FilterMethod   : Byte;
    InterlaceMethod: Byte;
  end;
var
  Pattern  : array[0..3] of Char;
  Idx      : Longint;
  TestBuf  : TBuffer;
  CurKey1,
  CurKey2  : Word;
  KeyFound : Boolean;
  Hex1,
  Hex2     : string[2];
  BaseCRC,
  InCRC,
  CRCVal,
  TargetCRC: Longword;
begin
  if ParamRec.UseNames then
    WriteLn('    info: trying to find PNG encryption key by brute force');

  // The width and height values of the PNG files stored in the archives of "A
  // New Beginning" have been encrypted. The encryption method is basically
  // the same as the one used to obfuscate the file table structure, but the
  // encryption key varies from file to file based on the file name.
  // If the correct key could not be derived from the file name, using various
  // assumptions (and the fact that each chunk of a PNG file is protected with
  // a CRC) we can brute-force it in reasonable time:
  // - The probability of an image having a dimension larger than 65535 pixels
  //   is low, so we assume that half of the bytes we want to decrypt are zero
  //   anyway. Thus, there are only 4 encrypted bytes left.
  // - The decryption takes place with a string of hex characters, i.e. there
  //   are only 16 possible partial keys per encrypted byte. Together with the
  //   first assumption, we get a key space of 65536 possibilites. Testing each
  //   of these possibilities with a CRC check afterwards takes only fractions
  //   of a second on contemporary machines.
  Result := False;

  if Length(Buffer) < SizeOf(TPNG_IHDR) + SizeOf(TargetCRC) then
  begin
    WriteLn('    warning: data too short for a PNG file');
    Exit
  end;

  // search the IHDR chunk
  Pattern := 'IHDR';
  Idx := SearchMem(SearchMethod_KnuthMorrisPratt, @Pattern[0], @Buffer[0],
    Length(Pattern), Length(Buffer));

  if (Idx < 0) or (Idx + SizeOf(TPNG_IHDR) + SizeOf(TargetCRC) >
    Length(Buffer)) then
  begin
    WriteLn('    warning: no valid IHDR found');
    Exit
  end;

  // get the targetted CRC value of the IHDR chunk
  Move(Buffer[Idx + SizeOf(TPNG_IHDR)], TargetCRC, SizeOf(TargetCRC));
  SwitchEndian(TargetCRC, True);

  SetLength(TestBuf, SizeOf(TPNG_IHDR));
  Move(Buffer[Idx], TestBuf[0], Length(TestBuf));

  // we assume that no image has a dimension larger than 65535
  TestBuf[4] := $00;
  TestBuf[5] := $00;
  TestBuf[8] := $00;
  TestBuf[9] := $00;

  // precompute the CRC value for as many fixed bytes as possible
  BaseCRC := $ffffffff;
  CalcCRC32(@TestBuf[0], 6, BaseCRC);

  // test all assumed key combinations
  KeyFound := False;
  for CurKey1 := 0 to 255 do
  begin
    // generate hex characters from the first part of the candidate key
    Hex1 := IntToHex(CurKey1, 2);

    // convert to lower case if necessary
    if Ord(Hex1[1]) >= $41 then
      Hex1[1] := Chr(Ord(Hex1[1]) + $20);
    if Ord(Hex1[2]) >= $41 then
      Hex1[2] := Chr(Ord(Hex1[2]) + $20);

    // try the first part of the decryption
    TestBuf[6] := TestBuf[6] xor Ord(Hex1[1]);
    TestBuf[7] := TestBuf[7] xor Ord(Hex1[2]);

    // compute the intermediate CRC
    InCRC := BaseCRC;
    CalcCRC32(@TestBuf[6], 4, InCRC);

    // search for second part of the key
    for CurKey2 := 0 to 255 do
    begin
      // generate hex characters from the second part of the candidate key
      Hex2 := IntToHex(CurKey2, 2);

      if Ord(Hex2[1]) >= $41 then
        Hex2[1] := Chr(Ord(Hex2[1]) + $20);
      if Ord(Hex2[2]) >= $41 then
        Hex2[2] := Chr(Ord(Hex2[2]) + $20);

      // try the second part of the decryption ...
      TestBuf[10] := TestBuf[10] xor Ord(Hex2[1]);
      TestBuf[11] := TestBuf[11] xor Ord(Hex2[2]);

      // ... and validate the chunk checksum
      CRCVal := InCRC;
      CalcCRC32(@TestBuf[10], Length(TestBuf) - 10, CRCVal);
      CRCVal := not CRCVal;

      // check whether the correct key has been found
      if CRCVal = TargetCRC then
      begin
        KeyFound := True;
        Break
      end;

      // get the original data back for the next test
      Move(Buffer[Idx + 10], TestBuf[10], 2)
    end;

    if KeyFound then
      Break;

    // get the original data back for the next test
    Move(Buffer[Idx + 6], TestBuf[6], 2)
  end;

  // if a key has been found, the test buffer contains the decrypted data
  if KeyFound then
  begin
    Move(TestBuf[0], Buffer[Idx], Length(TestBuf));
    Result := True
  end
  else
    WriteLn('    warning: no decryption key found')
end;

procedure GenerateNames(const Buffer: TBuffer);
var
  CrLf,
  NameBeg,
  NameEnd : TBuffer;
  NameFile: file;
  i,
  BaseIdx,
  BegIdx,
  EndIdx  : Longint;
  Name    : AnsiString;
  NameList: TStringList;
begin
  // The XML file within the game's main VIS archive contains mapping
  // information between a file's original name and its location within the
  // game's archive collection. Since parsing the XML file would be a bit of an
  // overkill here, we're just using a heuristic pattern search, which seems to
  // be working pretty well. 
  NameList := TStringList.Create;
  NameList.CaseSensitive := True;
  NameList.Sorted := False;

  // set up search patterns
  SetLength(CrLf, 2);
  SetLength(NameBeg, 2);
  SetLength(NameEnd, 2);

  CrLf[0] := $0d;
  CrLf[1] := $0a;

  NameBeg[0] := Ord('=');
  NameBeg[1] := Ord('"');

  NameEnd[0] := Ord('#');
  NameEnd[1] := Ord('"');

  // open target name table file
  if not OpenOutFile(NameFile, ParamRec.NameSpec) then
  begin
    WriteLn('    error opening name file');

    SetLength(CrLf, 0);
    SetLength(NameBeg, 0);
    SetLength(NameEnd, 0);
    Exit
  end;

  // Search the text buffer from back to front: The terminating combination #"
  // is much more significant than the opening =".
  BaseIdx := High(Buffer);
  while BaseIdx >= 0 do
  begin
    // search for the end of a name
    EndIdx := SearchMemReverse(SearchMethod_KnuthMorrisPratt, @NameEnd[0],
      @Buffer[0], 2, BaseIdx + 1);

    if EndIdx < 0 then
      Break;

    // search for the beginning of the name
    BegIdx := SearchMemReverse(SearchMethod_KnuthMorrisPratt, @NameBeg[0],
      @Buffer[0], 1, EndIdx + 1);

    if BegIdx < 0 then
      Break;

    BaseIdx := BegIdx - 1;

    // store the name (which might contain UTF-8 characters)
    SetLength(Name, EndIdx - BegIdx - 1);
    Move(Buffer[BegIdx + 2], Name[1], Length(Name));
    NameList.Add(Utf8ToAnsi(Name));
  end;

  // sort the list to easily find duplicates
  NameList.Sort;

  // output all unique names
  for i := 0 to NameList.Count - 1 do
    if (i = 0) or (NameList[i] <> NameList[i - 1]) then
    begin
      {$I-}
      BlockWrite(NameFile, NameList[i][1], Length(NameList[i]));
      BlockWrite(NameFile, CrLf[0], 2);
      {$I+}
      if IOResult <> 0 then
      begin
        WriteLn('    error writing to name file');
        Break
      end
    end;

  SetLength(CrLf, 0);
  SetLength(NameBeg, 0);
  SetLength(NameEnd, 0);
  NameList.Free;
  CloseFile(NameFile)
end;

procedure ReadNames;
var
  Count: Longint;
  Names: TextFile;
begin
  AssignFile(Names, ParamRec.NameSpec);
  {$I-}
  Reset(Names);
  {$I+}
  if IOResult <> 0 then
  begin
    WriteLn('warning: cannot read name file');
    Exit
  end;

  // count lines
  Count := 0;
  while not Eof(Names) do
  begin
    Inc(Count);
    ReadLn(Names)
  end;

  // prepare table and read names
  SetLength(NameTable, Count);

  Count := 0;
  Reset(Names);
  while not Eof(Names) do
  begin
    ReadLn(Names, NameTable[Count]);
    Inc(Count)
  end;

  CloseFile(Names)
end;

procedure GetName(const VISName: AnsiString; const DirEntry: TDirEntry;
  EntryNum: Longint; out RealName, OutName: AnsiString);
var
  i      : Longint;
  FileExt,
  Group,
  Count,
  Mask   : AnsiString;
begin
  RealName := '';
  OutName := '';

  // if a name table has been loaded, try to get the file's real name
  if ParamRec.UseNames then
  begin
    // Each entry in the name table is marked with an identifier that
    // corresponds to the archive file extension:
    // .vs035, file number 12 --> #s#035#00012#
    FileExt := AnsiLowerCase(ExtractFileExt(VISName));

    if FileExt = '.vis' then
    begin
      // the main VIS archive has the special designation "#v#-01#"
      Group := 'v';
      Count := '-01';
    end
    else if Copy(FileExt, 1, 2) = '.v' then
    begin
      // scene or character files follow the standard designation
      Group := Copy(FileExt, 3, 1);
      Count := Copy(FileExt, 4, Length(FileExt) - 3);

      // special case for video files
      if Group = 'v' then
        Group := 'm'
    end
    else
      Group := '';

    if Group <> '' then
    begin
      Mask := Format('#%s#%s#%.5d#', [Group, Count, EntryNum]);

      for i := 0 to High(NameTable) do
        if AnsiContainsStr(NameTable[i], Mask) then
        begin
          RealName := Copy(NameTable[i], 1, Length(NameTable[i]) -
            Length(Mask));

          OutName := RealName;
          if ParamRec.Unique then
          begin
            FileExt := ExtractFileExt(OutName);
            OutName := Copy(OutName, 1, Length(OutName) - Length(FileExt)) +
              Mask + FileExt
          end;

          Break
        end
    end
  end;

  // if no friendly name could be found, generate a default
  if OutName = '' then
    with DirEntry do
    begin
      OutName := VISName + '_' + AnsiLowerCase(IntToHex(Offset, 8));

      if (Flags and Flag_MainXML) <> 0 then
        OutName := OutName + '.xml'
      else if (Flags and Flag_CryptPNG) <> 0 then
        OutName := OutName + '.png'
      else
        OutName := OutName + '.dat'
    end
end;

function DetectGameKey(var Dir: TBuffer): Boolean;
var
  i        : Longint;
  DirHeader: TDirHeader;
  DirFooter: TDirFooter;
begin
  GameIdx := -1;
  for i := 0 to High(GameConfigs) do
  begin
    // Decrypt directory table with candidate key.
    Decrypt(Dir, 0, Length(Dir), GameConfigs[i].Key);

    // Get header and footer identifiers ...
    Move(Dir[0], DirHeader, SizeOf(DirHeader));
    Move(Dir[Length(Dir) - SizeOf(DirFooter)], DirFooter, SizeOf(DirFooter));

    // ... and check for validity
    if (DirHeader.ID = DirHeaderID) and (DirFooter.ID = DirFooterID) then
    begin
      WriteLn('  Info: Assuming ' + GameConfigs[i].Name +
        ' data file.');

      GameIdx := i;
      Break
    end
    else
      // wrong key: undo previous decryption
      Decrypt(Dir, 0, Length(Dir), GameConfigs[i].Key)
  end;

  Result := (GameIdx <> -1)
end;

procedure ExtractFiles(var VISFile: file; const VISName: AnsiString;
  const VISHeader: TVISHeader);
var
  i              : Longint;
  BaseOfs,
  BufPos,
  UncompBufPos   : Longword;
  Dir            : TBuffer;
  DirEntries     : TDirEntries;
  CompBlockHeader: TCompBlockHeader;
  RealName,
  OutName        : AnsiString;
  Buffer,
  UncompBuf      : TBuffer;
  OutFile        : file;
begin
  // read directory table
  SetLength(Dir, SizeOf(TDirHeader) + VISHeader.NumFiles * SizeOf(TDirEntry) +
    SizeOf(TDirFooter));
  {$I-}
  BlockRead(VISFile, Dir[0], Length(Dir));
  {$I+}
  if IOResult <> 0 then
  begin
    WriteLn('  error reading structures');
    SetLength(Dir, 0);
    Exit
  end;

  BaseOfs := FilePos(VISFile);

  // Try to detect the game key using the directory table identifiers as
  // markers.
  if not DetectGameKey(Dir) then
  begin
    WriteLn('  error detecting key');
    SetLength(Dir, 0);
    Exit
  end;

  // copy directory data to internal structures
  SetLength(DirEntries, VISHeader.NumFiles);
  Move(Dir[SizeOf(TDirHeader)], DirEntries[0],
    VISHeader.NumFiles * SizeOf(TDirEntry));

  SetLength(Dir, 0);

  // process all files
  for i := 0 to High(DirEntries) do
    with DirEntries[i] do
    begin
      SwitchEndian(Offset);
      SwitchEndian(CompSize);
      SwitchEndian(UncompSize);
      SwitchEndian(Flags);

      // determine a name for this file
      GetName(VISName, DirEntries[i], i, RealName, OutName);

      // if a name table is about to be generated, we're only interested in the
      // main XML file
      if ParamRec.GenNames and ((Flags and Flag_MainXML) = 0) then
        Continue;

      WriteLn('  ', OutName);

      // read file data
      SetLength(Buffer, CompSize);
      {$I-}
      Seek(VISFile, BaseOfs + Offset);
      BlockRead(VISFile, Buffer[0], Length(Buffer));
      {$I+}
      if IOResult <> 0 then
      begin
        WriteLn('    read error');
        Continue
      end;

      // Special case: If the "new" main XML file is found compressed, we have
      // to do a block-wise decompression (and decryption, if necessary).
      if ((Flags and Flag_MainXML2) <> 0) and (CompSize <> UncompSize) then
      begin
        BufPos := 0;
        UncompBufPos := 0;

        SetLength(UncompBuf, UncompSize);

        while Longint(BufPos + SizeOf(CompBlockHeader)) <= High(Buffer) do
        begin
          // Get the compressed and uncompressed sizes of the next block.
          Move(Buffer[BufPos], CompBlockHeader, SizeOf(CompBlockHeader));
          Inc(BufPos, SizeOf(CompBlockHeader));

          // The sizes are stored as big endian values, so switch them.
          with CompBlockHeader do
          begin
            SwitchEndian(UncompSize, True);
            SwitchEndian(CompSize, True)
          end;

          // Check for the special marker designating the chain end.
          if CompBlockHeader.CompSize = CompChainEnd then
            Break;

          // If the file is decrypted, we have to decrypt the block content
          // first.
          if (Flags and Flag_CryptDefault) <> 0 then
            Decrypt(Buffer, BufPos, CompBlockHeader.CompSize,
              GameConfigs[GameIdx].Key);

          // Now decompress the block content into the output buffer.
          if not Decompress(Buffer, BufPos, CompBlockHeader.CompSize, UncompBuf,
            UncompBufPos, CompBlockHeader.UncompSize) then
          begin
            WriteLn('    decompression error');
            Break
          end;

          BufPos := BufPos + CompBlockHeader.CompSize;
          UncompBufPos := UncompBufPos + CompBlockHeader.UncompSize
        end;

        Buffer := UncompBuf
      end
      else
      begin
        // regular cases for the "old" main XML file and all other types of
        // files

        // decrypt the whole file if necessary
        if (Flags and Flag_CryptDefault) <> 0 then
          Decrypt(Buffer, 0, Length(Buffer), GameConfigs[GameIdx].Key);

        // correct the PNG header if necessary
        if (Flags and Flag_CryptPNG) <> 0 then
          if not ParamRec.UseNames or not DecryptPNG(Buffer, RealName) then
            DecryptPNGForce(Buffer);

        // decompress if necessary
        if CompSize <> UncompSize then
        begin
          SetLength(UncompBuf, UncompSize);
          if Decompress(Buffer, 0, Length(Buffer), UncompBuf, 0,
            Length(UncompBuf)) then
            Buffer := UncompBuf
          else
            WriteLn('    decompression error')
        end
      end;

      // if the name file is to be genereted, use the first XML file for this
      // task
      if ParamRec.GenNames and ((Flags and Flag_MainXML) <> 0) then
      begin
        GenerateNames(Buffer);
        Break;
      end;

      // open output file and write final data
      if not OpenOutFile(OutFile, OutName) then
      begin
        WriteLn('    error opening output file');
        Continue
      end;

      {$I-}
      BlockWrite(OutFile, Buffer[0], Length(Buffer));
      {$I+}
      if IOResult <> 0 then
        WriteLn('    write error');

      CloseFile(OutFile)
    end;

  SetLength(Buffer, 0);
  SetLength(UncompBuf, 0);
  SetLength(DirEntries, 0)
end;

procedure ScanFiles;
var
  FileName,
  Dir,
  VISName  : string;
  VISFile  : file;
  VISHeader: TVISHeader;
  SearchRec: TSearchRec;
begin
  // read name table if available
  if ParamRec.UseNames then
    ReadNames;

  // get the directory information
  Dir := ExtractFilePath(ParamRec.FileSpec);

  // search for all files matching the file specification
  if FindFirst(ParamRec.FileSpec, faAnyFile - faDirectory,
    SearchRec) = 0 then
  begin
    repeat
      FileName := SearchRec.Name;
      VISName := Dir + FileName;

      WriteLn(VISName);

      if OpenInFile(VISFile, VISName) then
      begin
        if ReadHeader(VISFile, VISHeader) then
          ExtractFiles(VISFile, FileName, VISHeader)
        else
          WriteLn('  not a valid VIS file')
      end
      else
        WriteLn('  not accessible')
    until FindNext(SearchRec) <> 0
  end
  else
    WriteLn(Format('No matches for "%s"', [ParamRec.FileSpec]));

  SysUtils.FindClose(SearchRec)
end;

function ProcessParams: Boolean;
var
  i         : Longint;
  Param,
  ParamLower: AnsiString;
begin
  Result := False;

  if ParamCount = 0 then
    Exit;

  with ParamRec do
  begin
    FileSpec := '';
    NameSpec := '';
    GenNames := False;
    UseNames := False;
    Unique   := False;

    for i := 1 to ParamCount do
    begin
      Param := ParamStr(i);
      ParamLower := AnsiLowerCase(Param);
      if Copy(ParamLower, 1, 15) = '/generatenames:' then
      begin
        GenNames := True;
        if NameSpec <> '' then
          Exit;

        NameSpec := Copy(Param, 16, Length(Param) - 15)
      end
      else if Copy(ParamLower, 1, 7) = '/names:' then
      begin
        UseNames := True;
        if NameSpec <> '' then
          Exit;

        NameSpec := Copy(Param, 8, Length(Param) - 7)
      end
      else if ParamLower = '/unique' then
        Unique := True
      else if FileSpec <> '' then
        Exit
      else
        FileSpec := Param
    end;

    if GenNames and UseNames then
      Result := False
    else if (GenNames or UseNames) and (NameSpec = '') then
      Result := False
    else
      Result := (FileSpec <> '')
  end
end;

begin
  SetLength(NameTable, 0);

  if ProcessParams then
  begin
    if not ReadConfig then
      WriteLn(Format('Warning: %s', [LastError]));

    ScanFiles;

    SetLength(NameTable, 0)
  end
  else
    WriteLn('Usage: VISExt <VISFiles> [/names:<NameFile> | /generateNames:<Na' +
      'meFile>] [/unique]')
end.
