{ File    : $URL: file:///D:/SVN/svnDelphi/trunk/A%20New%20Beginning/VISExt/src/backend/KeyCfg.pas $
  Author  : $Author: Deniz Özmen $
  Changed : $Date: 2012-07-13 01:08:17 +0200 (Fr, 13 Jul 2012) $
  Revision: $Rev: 57 $

  Game key configuration provider for Visionaire Engine VIS archives.
}
unit KeyCfg;

interface

const
  DefaultConfigFileName: AnsiString = 'vis.key';

type
  TGameConfig = record
     ID,
     Name,
     Key : AnsiString;
  end;

  TGameConfigs = array of TGameConfig;

var
  LastError   : AnsiString;
  GameConfigs : TGameConfigs;

function ReadConfig(const ConfigFileName: AnsiString): Boolean; overload;
function ReadConfig: Boolean; overload;


implementation

uses
  SysUtils;

procedure Split(var Str1: AnsiString; Index: Longint; out Str2: AnsiString);
begin
  Str2 := Copy(Str1, 1, Index - 1);
  Str1 := Copy(Str1, Index + 1, Length(Str1) - Index)
end;

function ReadConfig(const ConfigFileName: AnsiString): Boolean;
var
  ConfigFile: TextFile;
  Line      : AnsiString;
  p         : Longint;
begin
  Result := False;

  if ConfigFileName = '' then
  begin
    LastError := 'Key config file name not specified.';
    Exit
  end;

  FileMode := fmOpenRead;
  AssignFile(ConfigFile, ConfigFileName);

  {$I-}
  Reset(ConfigFile);
  {$I+}
  if IOResult <> 0 then
  begin
    LastError := Format('Could not open key config file "%s".',
      [ConfigFileName]);
    Exit
  end;

  SetLength(GameConfigs, 0);
  while not Eof(ConfigFile) do
  begin
     ReadLn(ConfigFile, Line);

     if Line <> '' then
       SetLength(GameConfigs, Length(GameConfigs) + 1)
     else
       Continue;

     // get game ID ("short name")
     p := Pos(';', Line);
     if p <= 0 then
     begin
       LastError := 'Key config file format invalid.';
       CloseFile(ConfigFile);
       Exit
     end
     else
       Split(Line, p, GameConfigs[High(GameConfigs)].ID);

     // get full name
     p := Pos(';', Line);
     if p <= 0 then
     begin
       LastError := 'Key config file format invalid.';
       CloseFile(ConfigFile);
       Exit
     end
     else
       Split(Line, p, GameConfigs[High(GameConfigs)].Name);

     // get key
     GameConfigs[High(GameConfigs)].Key := Line;
     if GameConfigs[High(GameConfigs)].Key = ''
       then
     begin
       LastError := 'Key config file format invalid.';
       CloseFile(ConfigFile);
       Exit
     end
  end;

  CloseFile(ConfigFile);

  LastError := '';
  Result := True
end;

function ReadConfig: Boolean;
begin
  Result := ReadConfig(ExpandUNCFileName(IncludeTrailingPathDelimiter(
    ExtractFilePath(ParamStr(0)))) + DefaultConfigFileName)
end;

initialization
  LastError := '';
  SetLength(GameConfigs, 0);

end.
